%%% Monte Carlo Dynamic Classifier tool - MCDC tool

% Copyright (C) 2014 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
% Floor, Boston, MA 02110-1301 USA.

%
% Aspect.m
%
classdef Aspect < handle

  properties
    
    % ログファイル名
    % 空文字（デフォルト値）の場合、ログファイルを生成しない（標準出力への出力のみ）
    LogFileName
    
    % 中間状態および最終状態の .mat ファイルのファイル名接頭辞
    % 生成されるファイルのファイル名は、<設定値>_<Iteration 番号>.mat となる。
    MatFileNamePrefix
    
    % 中間状態の .mat ファイルを保存する場合 true
    % デフォルト値は false（保存しない）
    SavesIntermediateMat
    
    % SavesIntermediateMat = true の場合、中間状態を保存する Iteration 間隔
    % デフォルト値は 10
    IntermediateMatInterval
    
    % 最終状態の .mat ファイルを保存する場合 true
    % デフォルト値は false（保存しない）
    SavesFinalMat
    
  end
  
  properties (SetAccess = private)
    Algorithm
    In
    Out
    Started
    LogFid
    IterationStarted
    IterationCount
    IterationT
  end
  
  methods (Access = public)
    
    function this = Aspect()
      this.LogFileName = '';
      this.MatFileNamePrefix = '';
      this.SavesIntermediateMat = false;
      this.IntermediateMatInterval = 10;
      this.SavesFinalMat = true;
      
      this.Algorithm = [];
      this.In = [];
      this.Out = [];
      this.Started = false;
      this.LogFid = -1;
      this.IterationStarted = false;
      this.IterationCount = 0;
      this.IterationT = 0;
    end
    
    function Start(this, algorithm, in, out)
      this.End();
      
      if (~strcmp(this.LogFileName, ''))
        this.LogFid = fopen(this.LogFileName, 'w');
      end
      this.Algorithm = algorithm;
      this.In = in;
      this.Out = out;
      this.IterationStarted = false;
      this.IterationCount = out.j;
      this.IterationT = 0;
      
      this.Started = true;
    end
    
    function End(this)
      this.Started = false;
      this.closeAspect();
    end
    
    function StartIteration(this)
      if (this.Started && ~this.IterationStarted)
        this.IterationStarted = true;
        this.IterationCount = this.IterationCount + 1;
        if (this.In.J == 0)
          this.Log('Iteration %d', this.IterationCount);
        else
          this.Log('Iteration %d / %d', this.IterationCount, this.In.J);
        end
        
        this.IterationT = tic;
      end
    end
    
    function EndIteration(this)
      if (this.Started && this.IterationStarted)
        time = toc(this.IterationT);
        this.Log('Elapsed time is %f seconds.', time);
        
        saves = ((this.SavesIntermediateMat ...
            && mod(this.IterationCount, this.IntermediateMatInterval) == 0) ...
            || (this.SavesFinalMat && this.IterationCount == this.In.J)) ...
            && ~strcmp(this.MatFileNamePrefix, '');
        if (saves)
          matFile = sprintf('%s_%d.mat', this.MatFileNamePrefix, this.IterationCount);
          MCDCMatFile.Save(matFile, this.Algorithm, this.In, this.Out);
        end
        props = properties(this.Out);
        for i = 1:size(props, 1)
          this.logVariable(props{i}, getfield(this.Out, props{i}));
        end
        
        this.IterationStarted = false;
      end
    end
    
    function result = Log(this, varargin)
      s = strjoin({sprintf('%s -', datestr(now, 'yyyy/mm/dd HH:MM:SS')), sprintf(varargin{1:nargin - 1})});
      disp(s);
      if (this.Started && this.LogFid ~= -1)
        fprintf(this.LogFid, s);
        fprintf(this.LogFid, '\n');
      end
    end
    
  end
  
  methods (Access = private)
    
    function logVariable(this, name, value)
      info = whos('value');
      this.Log('  %s: %d bytes', name, info.bytes);
    end
    
    function closeAspect(this)
      if (this.LogFid ~= -1)
        [n, p, m, e] = fopen(this.LogFid);
        if ~strcmp(n, '')
          fclose(this.LogFid);
          this.LogFid = -1;
        end
      end
    end
    
    function delete(this)
      this.closeAspect();
    end
    
  end
  
end
