%%% Monte Carlo Dynamic Classifier tool - MCDC tool

% Copyright (C) 2014 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
% Floor, Boston, MA 02110-1301 USA.

%
% GPPF.m
%
% input:
%   xGrids: 状態変数値のグリッドデータ（次元毎データのセル配列）
%   initialStateKernelGens: 状態変数次元毎の初期カーネル関数生成オブジェクト (KernelGenerator) セル配列
%   initialObsKernelGens: 観測変数次元毎の初期カーネル関数生成オブジェクト (KernelGenerator) セル配列
%   initialStateMeanFuncs: 状態変数次元毎の初期平均値関数セル配列
%   initialObsMeanFuncs: 観測変数次元毎の初期平均値関数セル配列
%   gridDimForGramMatrix: カーネル分散共分散行列の作成に利用する次元
%   splineHandle: spline 補間関数生成関数ハンドル
%   x0: 状態変数の初期値
%   xaux: 状態変数に対する時刻ごとの制御入力
%   u: 状態変数に対する時刻ごとの制御入力
%   z: 観測値の時系列データ
%   N: Particle Filter 粒子数
%   J: Iteratrion 回数
%   K: 状態と観測の時刻差
%   aspect: Aspect オブジェクト
%
% output:
%   SKP: Iteration 毎の、状態変数カーネル関数パラメータ
%   OKP: Iteration 毎の、観測変数カーネル関数パラメータ
%   FV: Iteration 毎の、各グリッドにおける状態関数値 (次元毎データの多次元配列)
%   GV: Iteration 毎の、各グリッドにおける観測関数値 (次元毎データの多次元配列)
%   XE: Iteration 毎の、状態値時間推移
%   YE: Iteration 毎の、観測値時間推移
%   loglik: 尤度の Iteration 推移
%

classdef GPPF < Algorithm

  methods (Access = public)
  
    function this = GPPF()
      this = this@Algorithm(@RBFKernelGeneratorStrategyChoice1, @MCDCStrategyChoice1_GPPF);
    end

    function Execute(this, in, out, aspect)
      xdim = in.XDim();
      ydim = in.YDim();
      T = in.T();
      
      this.stateKGenStrategy.SetInitialKernelGens(in.stateKernelGens);
      this.obsKGenStrategy.SetInitialKernelGens(in.obsKernelGens);

      gridData = GridData(in.xGrids);
      this.mcdcStrategy.SetGridData(gridData);
      
      % Start
      aspect.Start(this, in, out);
      cleanupObj = onCleanup(@() aspect.End());
      
      % Grid Check
      CheckGridTransformation(VectorValuedFunction(in.stateMeanFuncs), gridData, aspect);

      %%%
      %%% MCMC Iteration
      %%%

      for j = (out.j + 1):in.J
      
        aspect.StartIteration();
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 1: Sample kernel parameters
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        stateKernelGens = this.stateKGenStrategy.GetGenerator(j, in, out);
        obsKernelGens = this.obsKGenStrategy.GetGenerator(j, in, out);

        for i = 1:xdim
          stateKernels{i} = stateKernelGens{i}.Generate();
          aspect.Log('  StateKernel[%d]: %s', i, stateKernels{i}.ToString());
        end 
        for i = 1:ydim
          obsKernels{i} = obsKernelGens{i}.Generate();
          aspect.Log('  ObsKernel[%d]: %s', i, obsKernels{i}.ToString());
        end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 3: Construct gram matrix on grid points
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        aspect.Log('  Creating GramMatrix using all dims...');
        
        gramMatrixGrid = gridData.GetFlatMatrix();

        for i = 1:xdim
          GramMatState{i} = stateKernels{i}.CreateGramMatrix(gramMatrixGrid);
          aspect.Log('    StateKernel[%d] Done', i);
        end
        
        for i = 1:ydim
          GramMatObs{i} = obsKernels{i}.CreateGramMatrix(gramMatrixGrid);
          aspect.Log('    ObsKernel[%d] Done', i);
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 4: Draw realization of the GP surface
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        aspect.Log('  Drawing GP surface...');    

        StateEqnVals = this.DrawGPSurface( ...
            gridData, ...
            in.gridDimForGramMatrix, ...
            this.mcdcStrategy.StateMean(j, in, out), ...
            this.mcdcStrategy.StateSigma(j, GramMatState));
        ObsEqnVals = this.DrawGPSurface( ...
            gridData, ...
            in.gridDimForGramMatrix, ...
            this.mcdcStrategy.ObsMean(j, in, out), ...
            this.mcdcStrategy.ObsSigma(j, GramMatObs));
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 5: Construct a B3 spline approximation
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        FnState = VectorValuedFunction(ModelFunctions2(in.splineHandle, gridData, StateEqnVals));
        FnObs = VectorValuedFunction(ModelFunctions2(in.splineHandle, gridData, ObsEqnVals));

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Stage 6: Estimate using particle filter
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        q = 1;
        r = 0.1;
        
        aspect.Log('  Estimating using particle filter... (N=%d)', in.N);

        particleFilter = this.mcdcStrategy.GetParticleFilter();        
        pf = particleFilter(in.x0, in.xaux', in.u', in.z', FnState, FnObs, q, r, in.N, in.K);
        
        acceptanceLogProbability = this.mcdcStrategy.AcceptanceLogProbability( ...
            j, in, out, pf.Loglik, StateEqnVals, ObsEqnVals, GramMatState, GramMatObs);
        aspect.Log('  Acceptance log probability = %f', acceptanceLogProbability);

        if rand(1) < exp(acceptanceLogProbability)
          out.AddSKP(cell2mat(arrayfun(@(i) stateKernels{i}.Properties(), 1:xdim, 'UniformOutput', false))');
          out.AddOKP(cell2mat(arrayfun(@(i) obsKernels{i}.Properties(), 1:ydim, 'UniformOutput', false))');
          out.AddFV(StateEqnVals);
          out.AddGV(ObsEqnVals);
          ji = find(rand(1) < cumsum(pf.Weights(:, T - in.K)), 1, 'first');
          xe = reshape(pf.Particles(ji, :, :), size(pf.Particles, 2), size(pf.Particles, 3));
          out.AddXE(xe);
          if length(in.xaux) > 0
            out.AddYE(FnObs([xe in.xaux']));
          else
            out.AddYE(FnObs(xe));
          end
          out.AddLoglik(pf.Loglik);

          this.mcdcStrategy.SaveGramMatrices(GramMatState, GramMatObs);

          if j == 1
            aspect.Log('  logLH = %f', pf.Loglik);
          else
            aspect.Log('  logLH = %f, accepted', pf.Loglik);
          end
        else
          aspect.Log('  logLH = %f, rejected', pf.Loglik);
        end
        
        out.SetIDXAsLast(j);
        out.j = j;
        aspect.EndIteration();
      
      end
      
    end
  
    function result = DrawGPSurface(this, gridData, pivot, means, gramMatrices)
      dim = size(means, 1);
      result = zeros([dim gridData.CountPerDim]);
      indices = gridData.GetAllIndices();
      for d = 1:dim
        for i = 1:size(indices, 1)
          result(d, indices(i, :)) = mvnrnd(means(d, indices(i, :)), gramMatrices{d});
        end
      end
    end

    function result = CalcLogProbability(this, values, gridData, pivot, means, gramMatrices)
      dim = size(means, 1);
      result = 0;
      indices = gridData.GetAllIndices();
      for d = 1:dim
        for i = 1:size(indices, 1)
          x = values(d, indices(i, :));
          mu = means(d, indices(i, :));
          sigma = gramMatrices{d};
          eigen = eig(sigma);
          if min(eigen) < 0
            sigma = sigma - min(eigen) .* eye(size(sigma));
          end
          % result = result + LogMvnPdf(x, mu, sigma);
          result = result + log(mvnpdf(x, mu, sigma));
        end
      end
    end

  end

end
