%%% Monte Carlo Dynamic Classifier tool - MCDC tool

% Copyright (C) 2014 Gareth W. Peters, Tomoko Matsui, Francois Septier, and Akira Tamamori

% This program is free software; you can redistribute it and/or modify it under the terms
% of the GNU General Public License as published by the Free Software Foundation; either
% version 2 of the License, or (at your option) any later version.

% This program is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
% PARTICULAR PURPOSE.  See the GNU General Public License for more details.

% You should have received a copy of the GNU General Public License along with this
% program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
% Floor, Boston, MA 02110-1301 USA.

classdef ResamplingScheme < handle
  % RESAMPLINGSCHEME
  % Resampling Schemes
  
  % METHODS (Static)
  % TODO: 
  % 1.  All of the Resampling Schemes should check the input Weights to
  %     ensure that they are already normalised!
  % 2.  Utilise the overload the Multinomial function, since the CDF component
  %     is used by all other schemes
  methods(Static)

    % Function: Multinomial
    % Input:    Weights (Vector)
    % Output:   Selected Indices (Vector)
    % NOTES:
    function results = Multinomial(Weights, StartIndex)
      % Multinomial.
%       fprintf('ResamplingScheme.Multinomial\n');

      if(nargin == 2 && StartIndex > 1)
          startIndex = StartIndex;
      else
        startIndex = 1;
      end
      
      numberWeights = length(Weights);
      
      % The CDF of the Weights
      empiricalCDF = cumsum(Weights);
      
      % Initialise the results data structure
      results = NaN(numberWeights, 1);
      
      % Iterate through the Weights
      for i = startIndex:numberWeights
        % Find the first weight that is greater than the randomly generated
        % number
        randNum = rand;
        index = find(empiricalCDF > randNum, 1);
        
        results(i) = index;
      end
    end

    % Function: Residual
    % Input:    Weights (Vector)
    % Output:   Selected Indices (Vector)
    % NOTES:  
    function results = Residual(Weights)
      % Residual.      
%       fprintf('ResamplingScheme.Residual\n');
      
      numberWeights = length(Weights);
      
      % Initialise the results data structure
      results = NaN(numberWeights, 1);
      
      % The counts of each index that will be in the results
      tmp = numberWeights .* Weights;
      indexReplicationCount = floor(tmp);
      
      % The number of indices that will be replicated in the results
      residualCount = sum(indexReplicationCount);
      
      % The number of indices that will be drawn using the
      % ResamplingScheme.Multinomial method
      numberMultinomialResampledWeights = numberWeights - residualCount;
      
      % Modify the Weights
      Weights = (tmp - indexReplicationCount)/numberMultinomialResampledWeights;
      
      % Replicate the Weights according to the replcationCount
      indicesToReplicate = find(indexReplicationCount > 0);
      
      % The number of times each index should be replicated
      indexReplicationCount = indexReplicationCount(indicesToReplicate);
      
      % Replicate the indicesToReplicate indexReplicationCount times in the
      % results
      C = arrayfun(@(x, y) repmat(x, [1 y]), indicesToReplicate, ...
                                indexReplicationCount, 'UniformOutput', false);
      results(1:residualCount, :) = cell2mat(C);
      
      % Draw the remaining Weights using the ResamplingScheme.Multinomial method
      multinomialResults = ResamplingScheme.Multinomial(Weights, residualCount + 1);     
      multinomialSelectedIndices = find(multinomialResults > 0);
      
      % Add the results from the 
      results((residualCount + 1):end) = ...
                                multinomialResults(multinomialSelectedIndices);
    end

    % Function: Stratified
    % Input:    Weights (Vector)
    % Output:   Selected Indices (Vector)
    % NOTES: 
    function results = Stratified(Weights)
      % Stratified.
%       fprintf('ResamplingScheme.Stratified\n');
      
      numberWeights = length(Weights);
      
      % Initialise the results data structure
      results = NaN(numberWeights, 1);
      
      % The CDF of the Weights
      empiricalCDF = cumsum(Weights);
      
      % Generate thresholds
      thresholds = NaN(numberWeights, 1);
      for i = 1:numberWeights,
        randNum = rand;
        thresholds(i) = (randNum + (i-1))/numberWeights;
      end

      % 
      for i = 1:numberWeights
        % The first index of the empiricalCDF that exceeds the current threshold
        results(i) = find(empiricalCDF > thresholds(i), 1);
      end
    end

    % Function: Systematic
    % Input:    Weights (Vector)
    % Output:   Selected Indices (Vector)
    % NOTES:  
    function results = Systematic(Weights)
      % Systematic. 
%       fprintf('ResamplingScheme.Systematic\n');
      
      numberWeights = length(Weights);
      
      % Initialise the results data structure
      results = NaN(numberWeights, 1);
      
      % The CDF of the Weights
      empiricalCDF = cumsum(Weights);
  
      % Generate thresholds
      randNum = rand;
      thresholds = linspace(0, 1 - 1/numberWeights, numberWeights) + ...
                                                          randNum/numberWeights;

      for i = 1:numberWeights
        % The first index of the empiricalCDF that exceeds the current threshold
        results(i) = find(empiricalCDF > thresholds(i), 1);
      end
    end

    % Function: Get Handle
    % Input:    Option (Integer)
    % Output:   Function Handle (Function Handle)
    % NOTES:
    function value = GetHandle(Option)
      % GetHandle.

      % Return the ResamplingScheme. method corresponding to the Option
      switch Option
        case 1
%         fprintf('ResamplingScheme.Multinomial\n');
          value = @(x) ResamplingScheme.Multinomial(x);
        case 2
%         fprintf('ResamplingScheme.Residual\n');
          value = @(x) ResamplingScheme.Residual(x);
        case 3
%         fprintf('ResamplingScheme.Systematic\n');
          value = @(x) ResamplingScheme.Systematic(x);
        case 4
%         fprintf('ResamplingScheme.Stratified\n');
          value = @(x) ResamplingScheme.Stratified(x);
        otherwise
          usageString = '1 = Multinomial, 2 = Residual, 3 = Systematic, 4 = Stratifed';
          error('ResamplingScheme:GetHandle:InvalidOption', usageString);
      end      
    end
  end
end
